program test_rms2int_diff

  ! This is the program to invert interval velocity from rms velocity by irls.
  ! This inversion is done without any regularization.
  ! The irls operator is from Antoine.

  use sep

  ! IRLS without regularization, from Antoine
!!$  use solver_irls_mod
!!$  use irls_priv

  ! IRLS with regularization, params for data fitting and model styling are diff.
  use solver_irls_diff_mod

  ! Forward and adjoint operator
  use causint
  use igrad1

  use solver_reg_med_mod
  use step_iter_mod

  use cgstep_mod
  use norm_mod
  use plane_search_mod

  use solver_reg_nrm_mod   !General Norm
  use solver_reg_IRLS_mod  !IRLS

 use dix_inv_mod
!!$  use solver_reg_l2_mod
!!$  use solver_smp_l2_mod

  ! Other operators used as constrants
  use mask1
 
  implicit none

  real, allocatable                            :: rmsvel(:), intervel(:), data(:)
  real, pointer                                :: weight(:), err(:), res(:)
  integer                                      :: n1, niter, nfreq,nerr, stat, i
  real                                         :: o1, d1, eps, rbar, it
  character(20)                                :: label1

  logical, pointer                             :: mask(:)
  real                                         :: qntd_irls, qntm_irls
  real                                         :: qntd, qntm

  integer :: psiter      !maximum iterations allowed in plan search
  integer :: qt_period   !Allow quantile change after certain iterations
  logical :: L_irls
  integer :: max_overshot_cnt
  character(12)  ::str_norm_mod, str_norm_dat
  real, dimension(:), allocatable :: error
  logical :: wmed,iter

  real, dimension(:),allocatable :: r_e,Fr_e,OF,OF_dat,OF_mod   ! recorded fitting error in every step, used for error plot
  real :: d_e,Fd_e,OF0,OF_dat0,OF_mod0

  call sep_init()

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)

  call to_history  ("n1", n1)
  call to_history  ("o1", o1)
  call to_history  ("d1", d1)
  call to_history  ("label1","time(s)")
  call to_history  ("label2","vel(km/s)")

  call from_param("eps"  , eps  , 100.0)
  write(0,*) "eps=", eps

  call from_param("qntd_IRLS" , qntd_irls , .8)
  call from_param("qntm_IRLS" , qntm_irls , .8)
  write(0,*) "qnt_IRLS=", qntd_irls, qntm_irls
  call from_param ("max_overshot_shrink_cnt", max_overshot_cnt, 0)
  write(0,*) "max_overshot_cnt=", max_overshot_cnt
  call from_param ("qt_period",qt_period,1000); write(0,*) "qt_period=",qt_period

  call from_param ("niter", niter, 1); write(0,*) "niter=",niter

  call from_param ("qntd",qntd,0.8)
  call from_param ("qntm",qntm,0.8)
  call from_param ("nfreq",nfreq)
  call from_param ("psiter",psiter,1); write(0,*) "psiter=",psiter

  call from_param   ("norm_dat", str_norm_dat, 'l2')
  call from_param   ("norm_mod", str_norm_mod, 'hybrid')
  write(0,*) "norm dat/res=", str_norm_dat, str_norm_mod

  call from_param   ("wmed", wmed, .false.)
  call from_param   ("iter", iter, .false.)  !iter is logical value, used in weighted-median method
  write(0,*) "wmed,iter=", wmed, iter

  call from_param ("IRLS", L_irls, .false.)
  write(0,*) "IRLS = ", L_irls

  allocate(rmsvel(n1), intervel(n1), data(n1))
  allocate(mask(n1))

  call sep_read(rmsvel,'in')

  mask = .true.;  mask(1) = .false.
  intervel = 0.;  intervel(1) = rmsvel(1)**2

  call mask1_init(mask)

  do i = 1, n1
     data(i) = rmsvel(i) * rmsvel(i) * i
  end do
  
  call wt_init(n1)

  !Invoke the misc solvers
  if (L_irls) then
     write(0,*) "invoke IRLS solver"
     call solver_irls_diff(m=intervel, d=data, Fop=causint_lop, stepper=cgstep, nAop=n1, niter=niter, nfreq=nfreq, Jop=mask1_lop, m0=intervel, &
          eps=eps,  Aop=igrad1_lop, qntd=qntd_irls, qntm=qntm_irls,Wop=wt_dix, verb=.true.)
  else if (wmed .and. iter) then
     write(0,*) "invoke the Weighted Median solver"
     call solver_reg_med(m=intervel, d=data, Fop=causint_lop,Aop=igrad1_lop,Wop=wt_dix,stepper=iterstep,&
          & nAop=size(intervel), niter=niter,eps=eps, verb=.true.)
  else

     if (exist_file("mod_iter")) then
        call to_history("n1",n1,"mod_iter")
        call to_history("o1",o1,"mod_iter")
        call to_history    ('d1',d1,"mod_iter")
        call to_history    ('n2',niter,"mod_iter")
        call to_history  ("label1","time(s)","mod_iter")
        call to_history  ("label2","vel(km/s)",'mod_iter')
     end if

     allocate (r_e(psiter*niter),OF(psiter*niter),OF_dat(psiter*niter),OF_mod(psiter*niter))
     allocate (Fr_e(niter))
     r_e = 0; Fr_e = 0; OF = 0; OF_dat = 0; OF_mod = 0
     d_e = 0; Fd_e = 0; OF0= 0; OF_dat0= 0; OF_mod0= 0

     write(0,*) "invoke the norm solver"
     call solver_reg_nrm(m=intervel, d=data, Fop=causint_lop, Aop=igrad1_lop,Wop=wt_dix, stepper=cgstep_ps, nAop=size(intervel), niter=niter, eps=eps, &
          & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, qnt_dat_in=qntd, qnt_mod_in=qntm, &
          & qt_period_in=qt_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt)
     write(0,*) "slover finished"
     stat = reg_nrm_get_error_stat(OF0,OF,OF_dat0,OF_dat,OF_mod0,OF_mod,d_e,r_e,Fd_e,Fr_e)
     write(0,*) "output the objective function value=", OF0, OF

     call solver_reg_nrm_close()

     deallocate (r_e,OF,OF_dat,OF_mod)
     deallocate (Fr_e)
  end if

  ! Output result
  intervel = sqrt(intervel)
  call sep_write(intervel)


  if (exist_file("recons")) then
     write(0,*) "output recon"
     call to_history("n1",n1,"recons")
     call to_history("o1",o1,"recons")
     call to_history("d1",d1,"recons")
     call to_history  ("label1","time(s)","recons")
     call to_history  ("label2","vel(km/s)",'recons')
     intervel = intervel**2
     stat = causint_lop(.false.,.false.,intervel,data)
     do i = 1, n1
        data(i) = sqrt(data(i)/i)
     end do
     call sep_write(data,"recons")
  end if

  deallocate(rmsvel, intervel, data)

  call sep_close()

end program test_rms2int_diff
