program GenSpikeMod

  use sep
  !use util_axisIO
  use random_mod
  implicit none

  integer ::spike_num,nt,nx
  real    ::dt,dx
  !type(myaxis) ::axis1,axis2

  integer ::max_spike_size_x,max_spike_size_t,size_x,size_t
  integer ::pos_x,pos_t
  integer ::seed
  integer ::i
  real, dimension(:), allocatable ::Ar_posx,Ar_post,Ar_sz_x,Ar_sz_t
  real, dimension(:), allocatable ::Ar_rand
  integer bg,end

  real,parameter  :: max_amp = 2.0
  real  :: gauge_t,gauge_x
  real, dimension(:,:), allocatable ::model

  integer:: ix_start,ix_end,it_start,it_end,hlf_x,hlf_t
  real :: dist_max,dist
  real :: it,ix

  gauge_x = 2.0; gauge_t = 1.0
  gauge_t = gauge_t**2; gauge_x = gauge_x**2

  call sep_init()
  call from_param("nt",nt)
  call from_param("nx",nx)
  call from_param("dt",dt)
  call from_param("dx",dx)

  call from_param("n_spike",spike_num)
  call from_param("max_spike_sz_x",max_spike_size_x)
  call from_param("max_spike_sz_t",max_spike_size_t)
  call from_param("rand_seed",seed)


  call random_init(seed)

  !call set_axis(axis1,nt,dt,0.0,"time(sec)")
  !call set_axis(axis2,nx,dx,0.0,"midpt(m)")
  !call print_axis(axis1)
  !call print_axis(axis2)

  allocate(Ar_posx(spike_num),Ar_post(spike_num),Ar_sz_x(spike_num),Ar_sz_t(spike_num),Ar_rand(4*spike_num))
  allocate(model(nt,nx))


  call rand_real_array_1d(Ar_rand)
  bg = 1; end = spike_num
  Ar_posx = Ar_rand(bg:end)
  bg = bg + spike_num; end = end + spike_num
  Ar_post = Ar_rand(bg:end)
  bg = bg + spike_num; end = end + spike_num
  Ar_sz_x = Ar_rand(bg:end)
  bg = bg + spike_num; end = end + spike_num
  Ar_sz_t = Ar_rand(bg:end)

  Ar_posx = Ar_posx * (nx-1) + 1.5
  Ar_post = Ar_post * (nt-1) + 1.5
  Ar_sz_x = Ar_sz_x * (max_spike_size_x-1) + 1.5
  Ar_sz_t = Ar_sz_t * (max_spike_size_t-1) + 1.5

  !Generate random spikes
  model = 0.0
  do i=1,spike_num
     !real 2 int
     pos_x = Ar_posx(i)
     pos_t = Ar_post(i)
     size_x = Ar_sz_x(i)
     size_t = Ar_sz_t(i)

 	if (mod(spike_num,3)==0 .or. pos_t<nt/4) then
	pos_x = (pos_x/16) * 16
	if (pos_x < 1) then
	pos_x=1
	end if
	end if

     !Place the reflectors
     hlf_x = size_x/2
     hlf_t = size_t/2
     ix_start = pos_x-hlf_x
     ix_end   = pos_x+hlf_x
     it_start = pos_t-hlf_t
     it_end   = pos_t+hlf_t

     ix_start = max(ix_start,1)
     ix_end   = min(ix_end,nx)
     it_start = max(it_start,2) !we dont like the reflector on the surface 
     it_end   = min(it_end,nt-1)!we dont like the reflector at the very bottom

     
     if (size_x==1 .and. size_t==1) then
	!write(0,*) "ix=", ix
        it = pos_t; ix = pos_x
        model(it,ix) = model(it,ix) + max_amp
     else
        dist_max = gauge_t*(hlf_t**2) + gauge_x*(hlf_x**2)
        do it=it_start,it_end
           do ix=ix_start,ix_end
              dist = gauge_t*((it-pos_t)**2) + gauge_x*((ix-pos_x)**2)

              model(it,ix) = model(it,ix) + &
                   & max_amp*exp(-4.0*(dist/dist_max))
           end do
        end do
     end if
  end do

  !call houtaxis(axis1,1)
  !call houtaxis(axis2,2)

  call to_history("n1",nt)
 call to_history("o1", 0)
 call to_history("d1", dt)
 call to_history("n2", nx)
 call to_history("o2", 0)
 call to_history("d2", dx)

 call to_history("label1", "time(sec)")
 call to_history("label2", "midpt(m)" )


  call sep_write(model)
	
  deallocate(Ar_posx,Ar_post,Ar_sz_x,Ar_sz_t)

  call sep_close()
end program GenSpikeMod
