! Kirchhoff modeling and migration.
program KMig

  use Kirchfast_subsamp_mod
  use sep

  use ddot_mod

  use solver_smp_l2_mod
  use cgstep_mod
  use solver_reg_med_mod
  use step_iter_mod
  use solver_smp_l1_mod
  use step_l1_mod

  use norm_mod
  use plane_search_mod
  use solver_reg_nrm_mod   !General Norm

  use solver_reg_IRLS_mod  !IRLS
  implicit none

  integer n1,n2,n3, iangle,nt,nx_d,nx_m
  ! <in.H Kirchhoff adj= vel=vel.H > out.H


  real, allocatable, dimension (:) :: model
  real, allocatable, dimension (:) :: data
  real, allocatable, dimension (:) :: vrms

  real, allocatable, dimension (:,:) :: data_full

  real                             :: thres_d,thres_m !For IRLS
  integer                          :: nfreq, nerr   !For IRLS
  integer                          :: niter
  real                             :: qnt_d,qnt_m
  logical :: L_irls

  integer :: max_overshot_cnt
  !|r|, |d|
  !|F'C'(r), F'C'(d)|
  real, dimension(:),allocatable :: r_e,Fr_e,OF,OF_dat,OF_mod   ! recorded fitting error in every step, used for error plot
  real :: d_e,Fd_e,OF0,OF_dat0,OF_mod0

  real, dimension(:),allocatable :: tmp_ar
  integer :: psiter      !maximum iterations allowed in plan search
  integer :: qt_period   !Allow quantile change after certain iterations

  real velocity, t0,dt,dx_d,dx_m,dx,d3, angle
  real    :: eps

  logical adj,inv,wmed,iter
  character(12)  ::str_norm_mod, str_norm_dat
  integer        ::stat

  integer :: sub_ratio
  integer ::offset, offset_c, ix_d

  real, dimension(:), allocatable :: error

  !Params
  call sep_init()

  call from_history('n1',n1)
  call from_history('n2',n2)
  nt = n1; 


  call from_param   ("iangle",iangle, 45)
  call from_param   ("eps",   eps,    100.0)
  write(0,*) "eps=", eps

  call from_history  ("o1",t0, 0.0)
  call from_history  ("d1",dt, 0.1)
  call from_history  ("d2",dx, 0.1)

  call from_param   ("adj",adj,.false.)
  call from_param   ("inv",inv,.false.)

  if (inv) then
     call from_param ("niter", niter, 1); write(0,*) "niter=",niter
     call from_param ("tresh_d",thres_d); 
     call from_param ("tresh_m",thres_m); 
     call from_param ("nfreq",nfreq)
     call from_param ("psiter",psiter,1); write(0,*) "psiter=",psiter

     call from_param ("IRLS", L_irls, .false.)
     write(0,*) "IRLS = ", L_irls

     call from_param ("qt_period",qt_period,1); write(0,*) "qt_period=",qt_period
     call from_param ("max_overshot_shrink_cnt", max_overshot_cnt, 0)
     write(0,*) "max_overshot_cnt=", max_overshot_cnt
  end if

  call from_param   ("qnt_d", qnt_d, .9) 
  call from_param   ("qnt_m", qnt_m, .9) 
  write(0,*) "quantile d/m = ", qnt_d,qnt_m

  call from_param   ("norm_dat", str_norm_dat, 'l2')
  call from_param   ("norm_mod", str_norm_mod, 'hybrid')
  write(0,*) "norm dat/res=", str_norm_dat, str_norm_mod

  call from_param   ("wmed", wmed, .false.)
  call from_param   ("iter", iter, .false.)  !iter is logical value, used in weighted-median method
  write(0,*) "wmed,iter=", wmed, iter

  call from_param   ("sub_ratio",sub_ratio,1)


  call to_history ("label1","time(sec)")
  call to_history ("label2","midpoint(m)")
  call to_history ("label3","Amp")

  call to_history    ("o1",t0)
  call to_history    ('d1',dt)
  call to_history    ('n1',n1)
  
  if (inv .or. adj) then
     !output the axis information
     nx_d=n2; dx_d = dx; nx_m = nx_d*sub_ratio; dx_m=dx_d/sub_ratio
     call to_history    ('n2',nx_m)
     call to_history    ('d2',dx_m)
  else !Forward modeling
     nx_m=n2; dx_m = dx; nx_d = nx_m/sub_ratio; dx_d=dx_m*sub_ratio
     call to_history    ('n2',nx_d)
     call to_history    ('d2',dx_d)
  end if

  write(0,*) 'allocate stuff', nt, nx_m, nt, nx_d, nt
  allocate (model(nt*nx_m))
  allocate (data(nt*nx_d))  
  allocate (vrms(nt))
  vrms=0;data=0; model=0
  call sep_read(vrms,'vel')
  vrms = vrms/2.0

  write(0,*) 'call the kirchfast init'
  call kirchfast_s_init(vrms,iangle,t0,dt,dx_d,dx_m,nt,nx_d,nx_m)

  if (adj) then !do the adjoint
     write(0,*) "Apply the adjoint only"
     call sep_read (data)
     stat = kirchfast_s_lop(adj,.false.,model,data)
     call sep_write (model)

  else if (inv) then !do the inversion with different scheme
     !These terms only applies to solver_smp_nrm subroutine
     
     write(0,*) 'call the sep_read'
     allocate (error (niter))
     call sep_read (data)
     write(0,*) "sep_read finished"

     if (L_irls) then
        write(0,*) "invoke IRLS solver"
        call IRLS_reg_init(thres_d,thres_m)
        call IRLS_reg(m=model,d=data,Fop=kirchfast_s_lop, Aop=unit_lop, stepper=cgstep, nAop=size(model), niter=niter, nfreq=nfreq, eps=eps,&
             & wght_d=weight_IRLS_d,wght_m=weight_IRLS_m,err=error)
     else
        if (wmed .and. iter) then
           write(0,*) "invoke the Weighted Median solver"
           call solver_reg_med(m=model, d=data, Fop=kirchfast_s_lop,Aop=unit_lop,stepper=iterstep,&
                & nAop=size(model), niter=niter,eps=eps, verb=.false.)
        elseif(wmed) then
           write(0,*) "invoke the l1solver_smp solver"
           call l1solver_smp(m=model, d=data, Fop=kirchfast_s_lop, stepper=l1step, niter=niter,verb=.false.)
        else

           allocate (r_e(psiter*niter),OF(psiter*niter),OF_dat(psiter*niter),OF_mod(psiter*niter))
           allocate (Fr_e(niter))
           r_e = 0; Fr_e = 0; OF = 0; OF_dat = 0; OF_mod = 0
           d_e = 0; Fd_e = 0; OF0= 0; OF_dat0= 0; OF_mod0= 0
           write(0,*) "invoke the norm solver"

           if (exist_file("mod_iter")) then
              call to_history("n1",nt,"mod_iter")
              call to_history("o1",t0,"mod_iter")
              call to_history    ('n2',nx_m,"mod_iter")
              call to_history    ('d2',dx_m,"mod_iter")
              call to_history    ('o3',0,"mod_iter")
              call to_history    ('d3',1,"mod_iter")
              call to_history    ('n3',niter,"mod_iter")

              call to_history ("label1","time(sec)","mod_iter")
              call to_history ("label2","midpoint(m)","mod_iter")
              call to_history ("label3","Amplitude","mod_iter")
           end if

           call solver_reg_nrm(m=model, d=data, Fop=kirchfast_s_lop, Aop=unit_lop, stepper=cgstep_ps, nAop=size(model), niter=niter, eps=eps, &
                & norm_dat_in=str_norm_dat, norm_mod_in=str_norm_mod, qnt_dat_in=qnt_d, qnt_mod_in=qnt_m,qt_period_in=qt_period, psiter=psiter, overshot_shrink_max_cnt_in=max_overshot_cnt)
           write(0,*) "slover finished"

           !Get error Statistics
           !function smp_reg_nrm_get_error_stat(OF0,OF,OF_dat_res0,OF_dat_res,OF_mod_res0,OF_mod_res,d,r,Fd,Fr) 
           stat = reg_nrm_get_error_stat(OF0,OF,OF_dat0,OF_dat,OF_mod0,OF_mod,d_e,r_e,Fd_e,Fr_e)

           call solver_reg_nrm_close()

           !Compute 1-|r|/|d|
           if (exist_file("avg_r")) then
              write(0,*) "output the avg_r, avg_r0=",d_e
              call to_history("n1",psiter*niter,"avg_r")
              call to_history("o1",1,"avg_r")
              call to_history("d1",1,"avg_r")
              call to_history("avg_r0",d_e,"avg_r")
              allocate(tmp_ar(niter*psiter))
              tmp_ar = 1 - r_e/d_e
              write(0,*) "r_e=",r_e, "d_e=",d_e
              call sep_write(tmp_ar,"avg_r")
              deallocate(tmp_ar)
           end if

           !Compute 1-|F'C'(r)|/|F'C'(d)|
           if (exist_file("meas_g")) then
              write(0,*) "output the meas_g,meas_g0=",Fd_e

              call to_history("n1",niter,"meas_g")
              call to_history("o1",1,"meas_g")
              call to_history("d1",1,"meas_g")
              call to_history("meas_g0",Fd_e,"meas_g")
              allocate(tmp_ar(niter))
              tmp_ar = 1 - Fr_e/Fd_e
              write(0,*) "Fr_e=",Fr_e, "Fd_e=",Fd_e
              call sep_write(tmp_ar,"meas_g")
              deallocate(tmp_ar)
           end if

           !Output the Objective Function value
           if (exist_file("of")) then
!               write(0,*) "output the objective function value=", OF0, OF
              call to_history("n1",niter*psiter,"of")
              call to_history("o1",1,"of")
              call to_history("d1",1,"of")
              call to_history("n2",3,'of')

              call to_history("objective_func_total_0",OF0,"of")
              call to_history("objective_func_dat_0",OF_dat0,"of")
              call to_history("objective_func_dat_0",OF_mod0,"of")
              call sep_write(OF,"of")
              call sep_write(OF_dat,"of")
              call sep_write(OF_mod,"of")
           end if

           deallocate (r_e,OF,OF_dat,OF_mod)
           deallocate (Fr_e)
        end if
     end if
     call sep_write (model)

     if (exist_file("recons")) then
        call to_history("n1",nt,"recons")
        call to_history("o1",t0,"recons")
        call to_history    ('n2',nx_m,"recons")
        call to_history    ('d2',dx_m,"recons")

        call to_history ("label1","time(sec)","recons")
        call to_history ("label2","midpoint(m)","recons")
        call to_history ("label3","Amp","recons")
	
	allocate(data_full(nt,nx_m))
	data_full=0
	!!Here in reconstructed version we choose to recon the finer record, e.g same res as in model space
        call kirchfast_lop2(.false.,model,data_full)
        call sep_write(data_full,"recons")
	deallocate(data_full)
     end if

     deallocate(error)
  else ! both adj and inv is false, do the forward operator
     write(0,*) "Do forward modeling"
     call sep_read (model)
     stat = kirchfast_s_lop (adj,.false.,model,data)
     call sep_write (data)

     !Output a interlaced version
     
       if (exist_file("data_interlaced")) then
        call to_history("n1",nt,"data_interlaced")
        call to_history("o1",t0,"data_interlaced")
        call to_history    ('n2',nx_m,"data_interlaced")
        call to_history    ('d2',dx_m,"data_interlaced")
        call to_history("n3",2,"data_interlaced")
        call to_history ("label1","time(sec)","data_interlaced")
        call to_history ("label2","midpoint(m)","data_interlaced")
        call to_history ("label3","Amp","data_interlaced")

	allocate(data_full(nt,nx_m))

	data_full=0
	offset = sub_ratio/2
	do ix_d=1,nx_d
		offset_c = (ix_d-1)*sub_ratio + offset
		data_full(:,offset_c) = data((ix_d-1)*nt+1:(ix_d)*nt)
	end do
	
        call sep_write(data_full,"data_interlaced")

	data_full=0
	call kirchfast_lop2(.false.,model,data_full)
	call sep_write(data_full,"data_interlaced")
	deallocate(data_full)
     end if
  end if

  deallocate (model,data,vrms)
  call sep_close()
  call kirchfast_s_close()
end program KMig

