program Test_veltran

  use sep
  use velsimp6
  use ddot_mod
  use solver_smp_nrm_mod
  use solver_smp_l2_mod
  use cgstep_mod
  use solver_smp_med_mod
  use step_iter_mod
  use solver_smp_l1_mod
  use step_l1_mod
  use solver_sep_mod !IRLS
  use norm_mod
  use plane_search_mod

  implicit none
  integer                          :: nt, nx, ns, iadj, n2, niter, stat
  real                             :: tresh, qnt
  integer                          :: nfreq, nerr
  real                             :: o1,d1, o2,d2, x0,dx, s0,ds
  real, dimension (:), allocatable :: data, model, error, recons
  logical                          :: adj,inv, wmed, iter
  character(len=12)                :: norm


  integer :: max_overshot_cnt
  !|r|, |d|
  !|F'C'(r), F'C'(d)|
  real, dimension(:),allocatable :: r_e,Fr_e,OF   ! recorded fitting error in every step, used for error plot
  real :: d_e,Fd_e,OF0

  real, dimension(:),allocatable :: tmp_ar
  integer :: psiter   !maximum iterations allowed in plan search
  integer :: qt_period!
  logical :: L_irls

  call sep_init ()                 ! start processing parameters

  call from_history (nt  , n2); !!Read in the first two parameters in the history
  call from_history ("o2", o2); call from_history ("d2", d2)

  call from_either  ("nx", nx, n2);  
  call from_either  ("ns", ns, 2*nx)  !n2, 2*nx are default value if read failed.
  call from_history ("o1", o1);       call from_history ("d1", d1)

  call from_either  ("x0", x0, o2);     call from_either  ("dx", dx, d2) 
  call from_either  ("s0", s0, o2);     call from_either  ("ds", ds, d2) 

  call from_param   ("adj",iadj,1) ; adj = (iadj == 1)
  call from_param   ("inv",inv,.false.)

  call from_param   ("qnt", qnt, .50) 
  write(0,*) "quantile = ", qnt

  call from_param   ("norm", norm, 'hybrid')
  call from_param   ("wmed", wmed, .false.)
  call from_param   ("iter", iter, .false.)  !iter is logical value, used in weighted-median method
  write(0,*) "wmed,iter=", wmed, iter

  !forward operator transform velscan(t,slowness) -> d(t,x), so the adjoint/inverse is doing what we want here

  if (adj .or. inv) then
     call to_history ("label2","slowness(sec/km)"); n2 = ns; o2 = s0; d2 = ds; 
  else
     call to_history ("label2","distance(km)"    ); n2 = nx; o2 = x0; d2 = dx; 
  end if

  call to_history ("o2", o2); call to_history ("d2", d2); 
  call to_history ("n2", n2)

  !if(adj.or.inv)then
  if (inv) then
     call from_param ("niter", niter, 1); write(0,*) "niter=",niter
     call from_param ("tresh",tresh); 
     call from_param ("nfreq",nfreq)
     call from_param ("psiter",psiter,1); write(0,*) "psiter=",psiter

     call from_param ("IRLS", L_irls, .false.)
     write(0,*) "IRLS = ", L_irls

     call from_param ("qt_period",qt_period,1); write(0,*) "qt_period=",qt_period
     call from_param ("max_overshot_shrink_cnt", max_overshot_cnt, 0)
     write(0,*) "max_overshot_cnt=", max_overshot_cnt

     !Add the error output part
     if (exist_file("err_d")) then
        call to_history("n1",psiter*niter,"err_d")
        call to_history("o1",1,"err_d")
        call to_history("d1",1,"err_d")
     end if
  endif

  if (exist_file ("error")) then
     call to_history ("n1",niter,"error")
  end if

  if (exist_file ("recons")) then
     call to_history ("o1", o1,"recons"); call to_history ("d1", d1,"recons"); call to_history ("n1", nt,"recons")
     call to_history ("o2", x0,"recons"); call to_history ("d2", dx,"recons"); call to_history ("n2", nx,"recons")
     call to_history ("label2","distance(km)","recons")
     allocate(recons(nt*nx))
  end if

  write(0,*) "parameter I/O finished"

  allocate (data(nt*nx), model(nt*ns))

  call velsimp6_init (o1,d1,x0,dx,s0,ds,nt,nx,ns)

  if (adj) then !do the adjoint
     write(0,*) "Apply the adjoint only"
     call sep_read (data)
     stat = velsimp6_lop (adj,.false.,model,data)
     call sep_write (model)

  else if (inv) then !do the inversion with different scheme
     nerr = niter/nfreq
     write(0,*) 'nerr', nerr
     allocate (error (niter))

     !These terms only applies to solver_smp_nrm subroutine
     allocate (r_e(psiter*niter),OF(psiter*niter))
     allocate (Fr_e(niter))
     r_e = 0; Fr_e = 0; OF = 0
     d_e = 0; Fd_e = 0; OF0= 0 

     call sep_read (data)
     write(0,*) "sep_read finished"
     if (L_irls) then
        write(0,*) "invoke IRLS solver"
        call IRLS_init(tresh)
        call IRLS(velsimp6_lop,cgstep,model,data,niter,nfreq,weight_IRLS,error)
     else 
        if (wmed .and. iter) then
           write(0,*) "invoke the Weighted Median solver"
           call solver_smp_med(m=model, d=data, Fop=velsimp6_lop, stepper=iterstep, niter=niter,verb=.true.)
        elseif(wmed) then
           write(0,*) "invoke the l1solver_smp solver"
           call l1solver_smp(m=model, d=data, Fop=velsimp6_lop, stepper=l1step, niter=niter,verb=.true.)
           write(0,*) "finish the l1solver_smp solver"
        else
           write(0,*) "invoke the hybrid norm solver"
           call solver_smp_nrm(m=model, d=data, Fop=velsimp6_lop, stepper=cgstep_ps, niter=niter,&
                & norm_in=norm, qnt=qnt,verb=.false.,qt_period_in=qt_period,psiter=psiter,err=error,&
                & overshot_shrink_max_cnt_in=max_overshot_cnt)
           write(0,*) "slover finished"

           !Get error Statistics
           !integer function smp_nrm_get_error_stat(OF0,OF,d,r,Fd,Fr)
           stat = smp_nrm_get_error_stat(OF0,OF,d_e,r_e,Fd_e,Fr_e)

           call solver_smp_nrm_close()

           !Compute 1-|r|/|d|
           if (exist_file("avg_r")) then
              write(0,*) "output the avg_r, avg_r0=",d_e
              call to_history("n1",psiter*niter,"avg_r")
              call to_history("o1",1,"avg_r")
              call to_history("d1",1,"avg_r")
              call to_history("avg_r0",d_e,"avg_r")
              allocate(tmp_ar(niter*psiter))
              tmp_ar = 1 - r_e/d_e
              write(0,*) "r_e=",r_e, "d_e=",d_e
              call sep_write(tmp_ar,"avg_r")
              deallocate(tmp_ar)
           end if

           !Compute 1-|F'C'(r)|/|F'C'(d)|
           if (exist_file("meas_g")) then
              write(0,*) "output the meas_g,meas_g0=",Fd_e

              call to_history("n1",niter,"meas_g")
              call to_history("o1",1,"meas_g")
              call to_history("d1",1,"meas_g")
              call to_history("meas_g0",Fd_e,"meas_g")
              allocate(tmp_ar(niter))
              tmp_ar = 1 - Fr_e/Fd_e
              write(0,*) "Fr_e=",Fr_e, "Fd_e=",Fd_e
              call sep_write(tmp_ar,"meas_g")
              deallocate(tmp_ar)
           end if

           !Output the Objective Function value
           if (exist_file("of")) then
              write(0,*) "output the objective function value=", OF0, OF
              call to_history("n1",niter*psiter,"of")
              call to_history("o1",1,"of")
              call to_history("d1",1,"of")
              call to_history("objective_func_value_0",OF0,"of")
              call sep_write(OF,"of")
           end if

        end if
     end if

     call sep_write (model)

     write(0,*) "output error"
     if (exist_file("error")) then
        call sep_write(error,"error")
     end if

  else ! both adj and inv is false, do the forward operator
     call sep_read (model)
     stat = velsimp6_lop (adj,.false.,model,data)
     call sep_write (data)
  end if

  if (inv .or. adj) then
     write(0,*) "output recon result"
     if (exist_file("recons")) then
        stat = velsimp6_lop (.false.,.false.,model,data)
        call sep_write(data,"recons")
        deallocate (recons); 
     end if
  end if

  deallocate (model,data)

  call sep_close()                ! end processing parameters

  if (allocated(r_e)) then
     deallocate(r_e,Fr_e,OF)
  end if

  if (allocated(error)) then
     deallocate(error)
  end if

  call exit (0)
end program Test_veltran
