module Kirchfast_subsamp_mod
  ! Kirchhoff migration and diffraction. (greased lightning)

  use adj_mod
  
  implicit none

  integer, private :: nt
  integer, private :: nx_d,nx_m  !nx_m should be multiple times of nx_d
  integer, private :: amax
  real, private :: t0,dt,dx_d,dx_m
  real, dimension(:), allocatable,save ::velhalf
  real, dimension(:), allocatable,save ::Ar_tmp
contains

  function unit_lop( adj,add,modl,data) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: modl,data 
    call adjnull (adj,add,modl,data )
    if (adj) then
       modl = modl + data
    else
       data = data + modl
    end if
    stat = 0
  end function unit_lop

  subroutine kirchfast_s_init(velhalf_in,amax_in,t0_in,dt_in,dx_d_in,dx_m_in,nt_in,nx_d_in,nx_m_in)
    real velhalf_in(nt_in)
    integer ::amax_in,nt_in,nx_d_in,nx_m_in
    real    ::t0_in,dt_in,dx_d_in,dx_m_in
    dx_d = dx_d_in
    nx_d = nx_d_in
    dx_m = dx_m_in
    nx_m = nx_m_in
    nt = nt_in
    t0 = t0_in
    dt = dt_in
    amax = amax_in

    allocate(velhalf(nt),Ar_tmp(nt*nx_m))
    velhalf = velhalf_in

  end subroutine kirchfast_s_init

  function kirchfast_s_lop( adj,add,modl,data) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: modl,data 
    call adjnull (adj,add,modl,data )
    !call kirchfast_s_lop2(adj,modl,data)
    if (adj) then
       Ar_tmp = 0
       call kirchfast_subsamp_lop(adj,Ar_tmp,data)
       call kirchfast_lop2(adj,modl,Ar_tmp)
    else
       Ar_tmp=0
       call kirchfast_lop2(adj,modl,Ar_tmp)
       call kirchfast_subsamp_lop(adj,Ar_tmp,data)
    end if
    stat = 0
  end function kirchfast_s_lop

  subroutine kirchfast_lop2(adj,modl,data)
    logical             :: adj
    integer 		:: ix,iz,it,ih,xstart, xend, h

    real  :: z,t,amp
    real  :: pi
    real  ::             tmp
    real  :: angle
    real  :: modl(nt,nx_m), data(nt,nx_m)
    integer :: ix_m
    pi = 2.0*ACOS(0.0)
    !amax 's unit is on degree, e.g if amax=45 means angle is 1/4 pi 

    !write(0,*) 'The amax sent in is ' , amax
    do ih= -nx_m, nx_m  !Use the finer grid, the grid in model space
       h = dx_m * ih ! h = offset
       do iz= 2, nt
          z = t0 + dt * (iz-1) ! z = travel-time depth
          t = sqrt( z**2 + (h/velhalf(iz))**2 )

          it = 1.5 + (t - t0) / dt
          if ( it > nt ) then
             exit
          end if
          !compute the angle

          if ( ih .ne. 0) then
             tmp = z/t
             angle = acos(z/t);
             angle = angle/pi * 180
             !write(0,*) 'tmp is ', tmp, 'angle is ', angle
             !throw away the points out of the aperture
             if (angle .gt. amax) then
                cycle
             endif
          endif

          amp = (z / t) * sqrt( nt*dt / t )
          xstart = 1-ih
          xstart = max0( 1, xstart)
          xend = nx_m-ih
          xend = min0( nx_m, xend)
          if ( adj == .false. ) then
             do ix= xstart, xend
		ix_m = ix+ih
                data(it,ix_m)=data(it,ix_m)+modl(iz,ix )*amp
             end do
          else
             do ix= xstart, xend
                ix_m = ix+ih
                modl(iz,ix) = modl(iz,ix )+data(it,ix_m)*amp
             end do
          end if
       end do
    end do

  end subroutine kirchfast_lop2

  subroutine kirchfast_subsamp_lop(adj,modl,data)
    logical             :: adj
    integer 		:: ix_d,ix1_m,ix2_m,it
    real ::modl(nt,nx_m), data(nt,nx_d)
    real ::wt1,wt2
    real,parameter ::eps=1e-6

    real ::x_d

    do ix_d=1,nx_d
       !Find the neighboring traces on finer grid ix_m
       x_d = (ix_d-1)*dx_d + 0.5*dx_d
       ix1_m = ((x_d)/dx_m -0.5+eps) + 1
       ix2_m = ix1_m+1
       !Interpolate wt
       wt2 = (x_d)/dx_m - (ix1_m-1+0.5)
       wt1 = 1.0 - wt2
       !write(0,*) 'wt1,wt2=', wt1,wt2
       !Intep
       if (adj) then
          modl(:,ix1_m) = modl(:,ix1_m) + wt1*data(:,ix_d)
          modl(:,ix2_m) = modl(:,ix2_m) + wt2*data(:,ix_d)
       else
          data(:,ix_d) = data(:,ix_d) + wt1*modl(:,ix1_m) + wt2*modl(:,ix2_m)
       endif
    end do

  end subroutine kirchfast_subsamp_lop

!!$  subroutine kirchfast_s_lop2(adj,modl,data)
!!$    logical             :: adj
!!$    integer 		:: ix,iz,it,ih,xstart, xend, h
!!$
!!$    real  :: z,t,amp
!!$    real  :: pi
!!$    real  ::             tmp
!!$    real  :: angle
!!$    real  :: modl(nt,nx_m), data(nt,nx_d)
!!$    integer :: ix_m,ix_d
!!$    pi = 2.0*ACOS(0.0)
!!$    !amax 's unit is on degree, e.g if amax=45 means angle is 1/4 pi 
!!$
!!$    !write(0,*) 'The amax sent in is ' , amax
!!$    do ih= -nx_m, nx_m  !Use the finer grid, the grid in model space
!!$       h = dx_m * ih ! h = offset
!!$       do iz= 2, nt
!!$          z = t0 + dt * (iz-1) ! z = travel-time depth
!!$          t = sqrt( z**2 + (h/velhalf(iz))**2 )
!!$
!!$          it = 1.5 + (t - t0) / dt
!!$          if ( it > nt ) then
!!$             exit
!!$          end if
!!$          !compute the angle
!!$
!!$          if ( ih .ne. 0) then
!!$             tmp = z/t
!!$             angle = acos(z/t);
!!$             angle = angle/pi * 180
!!$             !write(0,*) 'tmp is ', tmp, 'angle is ', angle
!!$             !throw away the points out of the aperture
!!$             if (angle .gt. amax) then
!!$                cycle
!!$             endif
!!$          endif
!!$
!!$          amp = (z / t) * sqrt( nt*dt / t )
!!$          xstart = 1-ih
!!$          xstart = max0( 1, xstart)
!!$          xend = nx_m-ih
!!$          xend = min0( nx_m, xend)
!!$          if ( adj == .false. ) then
!!$             do ix= xstart, xend
!!$		ix_m = ix+ih
!!$                ix_d = (((ix_m-1)*dx_m-0.5*dx_d)/dx_d+1.5)
!!$                data(it,ix_d)=data(it,ix_d)+modl(iz,ix )*amp
!!$             end do
!!$          else
!!$             do ix= xstart, xend
!!$                tmp_d = data(it,ix_d)
!!$		ix_m = ix+ih
!!$		ix_d = (((ix_m-1)*dx_m-0.5*dx_d)/dx_d+1.5)
!!$                modl(iz,ix) = modl(iz,ix )+data(it,ix_d)*amp
!!$             end do
!!$          end if
!!$       end do
!!$    end do
!!$
!!$  end subroutine kirchfast_s_lop2

  subroutine kirchfast_s_close()
    if (allocated(velhalf)) then
       deallocate(velhalf)
    end if
  end subroutine kirchfast_s_close

end module Kirchfast_subsamp_mod
