module solver_irls_diff_mod

! rbar is determined by the quantile of the model residuals

  use chain0_mod
  use irls
  use solver_report_mod
  use quantile_mod
  implicit none
  logical, parameter, private :: T = .true., F = .false.

contains

  subroutine solver_irls_diff(m, d, Fop, stepper, nAop, Jop, qntd, qntm, & 
                             niter, nfreq, eps, Aop, Wop, m0, mmov, rmov, err, verb)
    optional :: Aop, Wop, mmov, rmov, err, verb, Jop, m0
    interface
!------------Forward modeling operator-----------------
      integer function Fop(adj, add, m, d)
        real               :: m(:), d(:)
        logical,intent(in) :: adj, add   
      end function
      
!------------Regularization operator-------------------
      integer function Aop(adj, add, m, d)
        real               :: m(:), d(:)
        logical,intent(in) :: adj, add
      end function

!------------Weighting function operator---------------
      integer function Wop(adj, add, m, d)
        real               :: m(:), d(:)
        logical,intent(in) :: adj, add 
      end function

!-------------------Mask operator----------------------
      integer function Jop (adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function

!-----------------stepper (cgstep)----------------------
      integer function stepper(forget, m, g, rr, gg)
        real, dimension(:) :: m, g, rr, gg
        logical            :: forget 
      end function

    end interface

    real   , dimension(:),        intent(in)  :: d, m0
    integer,                      intent(in)  :: niter, nAop, nfreq

    real                                      :: eps
    logical,                      intent(in)  :: verb
    real   , dimension(:),        intent(out) :: m
    real   , dimension(:),        intent(out) :: err
    real   , dimension(:,:),      intent(out) :: mmov, rmov

    real   , dimension(size(m))               :: g
    real   , dimension(size(d)+nAop), target  :: rr, gg, tt
    real   , dimension(:), pointer            :: rd, gd, td
    real   , dimension(:), pointer            :: rm, gm, tm
    integer                                   :: iter, ifreq, stat, num, nud
    logical                                   :: forget
    real   , allocatable                      :: wtm(:), wtd(:)
    real   ,                      intent(in)  :: qntm, qntd
    real                                      :: rbarm, rbard
    
    rd => rr(1:size(d)); rm => rr(1+size(d):)
    gd => gg(1:size(d)); gm => gg(1+size(d):)
    td => tt(1:size(d)); tm => tt(1+size(d):)

    if(.not. allocated(wtm)) allocate(wtm(nAop))
    if(.not. allocated(wtd)) allocate(wtd(nAop))

    num = qntm * nAop
    nud = qntd * size(d)

! Initialization------------------------------------
    if (present(Wop)) then
        stat = Wop(F, F, -d, rd)
    else
        rd = -d
    end if
    ! Initial solution------------------------------
    if (present( m0)) then
      m=m0 !m = m0
      if (present( Wop)) then
        call chain0(Wop,Fop,F,T,m,rd,td)
      else
        stat = Fop(F,T,m,rd ) !Rd+= WF m0
      end if
    else
      m=0
    end if

    rm = 0.
    wtm = 1. 
    wtd = 1.
    forget = F
    ifreq = 0

    do iter=1, niter
!      restart
!      -------
       if (forget) then
          ifreq = ifreq + 1
		
	  if (iter < 2*(nfreq+1)) then !Only change the Rth at the beginning
          	rbarm = quantile(num, rm)
		rbard = quantile(nud, rd)
	end if
          wtm   = 1. / sqrt( sqrt (1. + (rm/rbarm)**2))
          rbard = quantile(nud, rd)
          wtd   = 1. / sqrt( sqrt (1. + (rd/rbard)**2))
!         reweight the residual
!         --------------------- 
          rm   = rm * wtm                                         ! rm = rm * Wdiag
          rd   = rd * wtd
       end if

!      compute the data gradient
       if (present(Wop)) then
          call chain0(Wop, Fop, T, F, g, rd * wtd, td)
       else
          stat = Fop(T, F, g, rd * wtd)                   ! g = (WF)' rd
       end if

!      compute the model gradient
       if (present(Aop)) then
          stat = Aop(T, T, g, eps*wtm*rm)                  ! g = g + eps * A' Wdiag' rm
       else
          g = g + eps*wtm*rm                               ! g = g + eps * Wdiag'rm
       end if

       if (present(Jop)) then
          tm=g
          stat = Jop(F,F,tm, g )
       end if

!      compute the perturbed data for the data fitting goal
!      ----------------------------------------------------
       if (present(Wop)) then
          call chain0(Wop, Fop, F, F, wtd*g, gd, td)
       else
          stat = Fop(F, F, wtd * g, gd)                   ! gd = WF g
       end if

!      compute the perturbed data for the model fitting goal
!      -----------------------------------------------------
       if (present(Aop)) then
          stat = Aop(F, F, eps*wtm*g, gm)                  ! gm = eps * Wdiag A g
       else
          gm = eps*wtm*g
       end if 
 
       stat = stepper(forget, m, g, rr, gg)

       if(stat == 1) exit

       forget = (mod(iter,nfreq)==0)                 ! periodic restart

       if (present(mmov)) mmov(:,iter) = m
       if (present(rmov)) rmov(:,iter) = rr(1:size(d))
       if (present(err))  err(iter)    = dot_product(rd,rd)+dot_product(rm,rm)
       if (present(verb)) write(0,*)"iter=",iter,"rd=",real(dot_product(rd,rd)),"rm=",real(dot_product(rm,rm))

    end do

    if(allocated(wtm))   deallocate(wtm)
    if(allocated(wtd))   deallocate(wtd)
    return
  end subroutine solver_irls_diff

end module solver_irls_diff_mod





