module solver_reg_IRLS_mod
! 0 = F m - d
! 0 = A m

        use cgstep_mod
        use chain0_mod 
        use ddot_mod
        use quantile_mod 
   
        implicit none
        logical, parameter, private     :: T = .true., F = .false.
        logical,            private     :: time
        real    , private               :: scale
        real    ,save                   :: tstart, tfinish, ttot, tinv, telapse, tleft, tnow, told
        integer ,save                   :: tarray(8), tarray1(8), tarray2(8), itniter, itleft
        real    ,save                   :: treshold_d, treshold_m !(threshold for data and model residue respectively

        character(len=256),private,save :: report_
        logical,            private     :: report_fil
        contains
!#################################irls-Regularized Yang

          subroutine IRLS_reg( m,d,Fop, Aop, stepper, nAop, niter, nfreq, eps, wght_d,wght_m,err,m0)
            optional ::        err,m0
            interface
               integer function wght_d( res, w)                  
                 real, dimension (:)   :: res, w  
               end function wght_d
               integer function wght_m( res, w)                  
                 real, dimension (:)  :: res, w  
               end function wght_m
               integer function Fop( adj, add, m, d)     
                 logical, intent (in)  :: adj, add
                 real, dimension (:)  :: m, d      
               end function Fop
               integer function Aop( adj, add, m, d)     
                 logical, intent (in)  :: adj, add
                 real, dimension (:)  :: m, d      
               end function Aop
               integer function stepper( forget, x, g, rr, gg)    
                 logical               :: forget  
                 real, dimension (:) :: x, g, rr, gg
               end function stepper
            end interface

            real, dimension (:), intent (in)  :: d                ! data
            real, dimension (:), intent (out) :: m                  ! solution
            real, dimension (:), intent (out) :: err                ! error
            integer,             intent (in)  :: niter, nfreq       ! iterations
            real, dimension (size (m))        :: g                  ! gradient
            real     ::  eps
            integer  ::  nAop

            integer                           :: i, stat
            logical                           :: forget
            real                              :: f0
            real, dimension(size(m)), intent(in)    :: m0

            real, dimension (size (d) + nAop), target :: rr, gg, tt,wt         ! res, CG, weight
            real, dimension(:), pointer                :: rd, gd, td
            real, dimension(:), pointer                :: rm, gm, tm
            real, dimension(:), pointer                :: wt_d,wt_m

            !Assign the array pointers
            rd => rr(1:size(d))
            rm => rr(1+size(d):)
            gd => gg(1:size(d))
            gm => gg(1+size(d):)
            td => tt(1:size(d))
            tm => tt(1+size(d):)
            wt_d=>wt(1:size(d))
            wt_m=>wt(1+size(d):)

            m = 0
            if (present(m0)) then
               m = m0
            end if

            wt = 1.
            forget = T                            ! initialize

            do i = 1, niter  
               if (forget) then                   !restart

                  rd = -d            
                  ! rd+=Fm0
                  stat = Fop(F,T,m,rd)

                  ! rm+=Am0
                  rm = 0
                  stat = Aop(F,T,eps*m,rm)

                  stat = wght_d( rd, wt_d)            ! compute weighting
                  stat = wght_m( rm, wt_m)

                  rr = rr * wt
               end if

               if (present(err)) then
                  if(i.eq.1) f0 = sum(dprod(rr,rr))
                  err(i) = sum(dprod(rr,rr))/f0
!                   write(0,*) 'residual(',i,')=',err(i)
               endif

               stat = Fop( T, F, g, wt*rr)            ! g  = F' W' rd + eA' rm
               stat = Aop( T, T, g, eps*rm)

               stat = Fop( F, F, g, gd)               ! gd = WFg
               stat = Aop( F, F, eps*g, gm)           ! gm = W*eps*A*g
               gg = gg * wt  

               stat = stepper( forget, m, g, rr, gg)      ! step in x and rr  
               forget = ( mod( i, nfreq) .eq. 0)       ! periodic restart                     
            end do
          end subroutine IRLS_reg
        !#################################irls weights    
        subroutine IRLS_reg_init (tr_d,tr_m)
          real   :: tr_d,tr_m
          treshold_d = tr_d
          treshold_m = tr_m
        end subroutine IRLS_reg_init

        function weight_IRLS_d(res,w) result(stat)
          integer                    ::stat,i
          real, dimension(:)         ::res,w  
          do i=1, size(res)
             w(i) = 1./sqrt(sqrt(1.+res(i)*res(i)/(treshold_d*treshold_d)))
          end do
          stat = 0
        end function weight_IRLS_d

        function weight_IRLS_m(res,w) result(stat)
          integer                    ::stat,i
          real, dimension(:)         ::res,w  
          do i=1, size(res)
             w(i) = 1./sqrt(sqrt(1.+res(i)*res(i)/(treshold_m*treshold_m)))
          end do
          stat = 0
        end function weight_IRLS_m

!!$        integer function l1 (res, weight)
!!$          real, dimension (:)  :: res, weight
!!$          real                 :: rbar
!!$          rbar = quantile( int( 0.5*size(res)), abs (res))            
!!$          weight = 1. / sqrt( sqrt (1. + (res/rbar)**2))
!!$          l1 = 0
!!$        end function l1
      end module solver_reg_IRLS_mod
