program test_rms2int

  ! This is the program to invert interval velocity from rms velocity 
  ! by L1-Iterative Plane-Search
  ! This inversion is done without any regularization.

  use solver_smp_nrm_mod
  use solver_reg_nrm_mod
  use cgnorm_mod


  ! use solver_prc_nrm_mod


  use causint
  use sep
  use mask1
  use igrad1
  use wght1


  implicit none
  
  real, allocatable              :: rmsvel(:), intervel(:), data(:)
  real, pointer                  :: err(:), res(:)
  real                           :: o1, d1, l1par, eps, it, qnt
  integer                        :: reg, prc
  integer                        :: n1, niter, nfreq, stat, nmem, i,psiter
  logical, pointer               :: mask(:) 
  character(20)                  :: label1,norm

  call sep_init(SOURCE)

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)

  call to_history  ("n1", n1)
  call to_history  ("o1", o1)
  call to_history  ("d1", d1)

  ! Horses Params
  call from_param("norm", norm , "hybrid")
  call from_param("qnt" , qnt , 0.90 )
  call from_param("psiter", psiter, 250); 


  ! General
  call from_param("reg"  , reg  , 0 )
  call from_param("eps"  , eps  , 0.1)
  call from_param("prc"  , prc  , 0)
  call from_param("niter", niter, n1*3+1); 


  allocate(rmsvel(n1), intervel(n1), data(n1))
  allocate(mask(n1))

  ! READ IN Vrms and Form Data
  call sreed('in', rmsvel, n1*4)
  open(unit=1, file='realdata.data')
  write(unit=1, fmt=*) rmsvel
  do i = 1, n1
     it = (i-1) * d1 + o1
     data(i) = rmsvel(i) * rmsvel(i) * i
  end do
  ! data(n1-10)=  data(n1-10) + 1000    ! ADD Single SPIKE


  ! Form Mask operator for element 1,N
  intervel=0.; intervel(1)=data(1)
  mask = .true.; 
  mask(1) = .false.
  ! mask(n1) = .false.
  call mask1_init(mask)
  write(0,*) " Vrm,Vint,Data (1)",rmsvel(1),intervel(1), data(1)

  
  if (prc==1) then
     !      call solver_prc_nrm(m=intervel,d=data,Fop=causint_lop,Sop=causint_lop,&
     !           stepper=cgnorm,nSop=n1,niter=niter,eps=eps,norm=norm,qnt=qnt,psiter=psiter,&
     !           Jop=mask1_lop,verb=.true.,p0=intervel)
  elseif(reg==1) then
     call solver_reg_nrm(m=intervel,d=data,Fop=causint_lop,Aop=igrad1_lop, &
          stepper=cgnorm,nAop=n1,niter=niter,eps=eps,& ! Wop=wght1_lop, & !! Wop:BAD end
          norm=norm,qnt=qnt,psiter=psiter,Jop=mask1_lop,verb=.true.,m0=intervel)
  else
     call solver_smp_nrm(m=intervel,d=data,Fop=causint_lop,&  ! Wop=wght1_lop, & !! Wop:BAD end
          stepper=cgnorm, niter=niter,norm=norm,qnt=qnt,psiter=psiter,Jop=mask1_lop,&
          verb=.true.,m0=intervel )
  end if


  ! WRITE OUT Vint
  intervel = sqrt(intervel)
  call srite("out",intervel,n1*4)
  write(0,*) "Vint(1)",intervel(1)

  deallocate(rmsvel, intervel, data, mask)
end program

