program Test_veltran

  use sep
  use velsimp6
  use ddot_mod
  use solver_smp_nrm_mod
  use cgnorm_mod
  use solver_smp_l2_mod
  use cgstep_mod
  use solver_smp_med_mod
  use step_iter_mod
  use solver_smp_l1_mod
  use step_l1_mod
!   use step_inner_mod

  implicit none
  integer                          :: nt, nx, ns, iadj, n2, niter, stat
  real                             :: tresh, qnt
  integer                          :: nfreq, nerr
  real                             :: o1,d1, o2,d2, x0,dx, s0,ds
  real, dimension (:), allocatable :: data, model, error, recons
  logical                          :: adj,inv, wmed, iter, inner
  character(len=10)                :: norm

  call sep_init ()                 ! start processing parameters
  
    call from_history (nt  , n2); call from_history ("o2", o2); call from_history ("d2", d2)
    call from_either  ("nx", nx, n2);   call from_either  ("ns", ns, 2*nx)
    call from_history ("o1", o1);       call from_history ("d1", d1)
    call from_either  ("x0", x0, o2);     call from_either  ("dx", dx, d2) 
    call from_either  ("s0", s0, o2);     call from_either  ("ds", ds, d2) 

    call from_param   ("adj",iadj,1) ; adj = (iadj == 1)
    call from_param   ("qnt", qnt, .50)
    call from_param   ("norm", norm, 'hybrid')
    call from_param   ("wmed", wmed, .false.)
    call from_param   ("iter", iter, .false.)
    call from_param   ("inner", inner, .false.)

 

   if (adj) then; call to_history ("label2","slowness(sec/km)"); n2 = ns; o2 = s0; d2 = ds; 
            else; call to_history ("label2","distance(km)"    ); n2 = nx; o2 = x0; d2 = dx; end if
    call to_history ("o2", o2); call to_history ("d2", d2); call to_history ("n2", n2)
    if(adj.or.inv)then
            call from_param ("niter", niter, 1); 
            call from_param ("tresh",tresh); 
            call from_param ("nfreq",nfreq)
    endif
    if (exist_file ("error")) then; call to_history ("n1",niter,"error"); end if
    if (exist_file ("recons")) then
        call to_history ("o1", o1,"recons"); call to_history ("d1", d1,"recons"); call to_history ("n1", nt,"recons")
        call to_history ("o2", x0,"recons"); call to_history ("d2", dx,"recons"); call to_history ("n2", nx,"recons")
        call to_history ("label2","distance(km)","recons")
        allocate(recons(nt*nx))
    end if
    call sep_close()                ! end processing parameters
    allocate (data(nt*nx), model(nt*ns))

    call velsimp6_init (o1,d1,x0,dx,s0,ds,nt,nx,ns)
    if (adj) then  
        nerr = niter/nfreq
        write(0,*) 'nerr', nerr
        allocate (error (niter))
        call sep_read (data)

        if (wmed .and. iter) then
           call solver_smp_med(m=model, d=data, Fop=velsimp6_lop, stepper=iterstep, niter=niter,verb=.true.)
!         if (wmed .and. inner) then
!            call sovler_smp_med(m=model, d=data, Fop=velsimp6_lop, stepper=innerstep, niter=niter,verb=.true.)
        elseif(wmed) then
           call l1solver_smp(m=model, d=data, Fop=velsimp6_lop, stepper=l1step, niter=niter,verb=.true.)
        else
           call solver_smp_nrm(m=model, d=data, Fop=velsimp6_lop, stepper=cgnorm, niter=niter, norm=norm, qnt=qnt,verb=.true.)
        end if
        call sep_write (model)
        if (exist_file("recons")) then; stat = velsimp6_lop (.false.,.false.,model,data); call sep_write(data,"recons"); deallocate (recons); end if
  else
        call sep_read (model)
        stat = velsimp6_lop (adj,.false.,model,data)
        call sep_write (data)
  end if
  deallocate (model,data)
  call exit (0)
end program Test_veltran
