module solver_reg_nrm_mod 
  ! 0 = W (F J m - d)
  ! 0 =    A   m

  use sep
  use chain0_mod 
  use solver_report_mod
  use cgnorm_mod

  implicit none
  logical, parameter, private  :: T = .true., F = .false.


  contains
  subroutine solver_reg_nrm( m,d, Fop, Aop, stepper, nAop, niter,eps   ,  &
    &           norm,qnt,psiter,Wop,Jop,m0,rm0,err,resd,resm,mmov,rmov,verb)
    optional :: norm,qnt,psiter,Wop,Jop,m0,rm0,err,resd,resm,mmov,rmov,verb
    interface  

      !-------------------- begin definitions --------------------
      integer function Fop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function Aop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function Wop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function Jop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function 
      integer function stepper(forget,m,g,rr,gg)
        real, dimension(:)  ::        m,g,rr,gg    
        logical             :: forget
      end function
    end interface 
    real, dimension(:),    intent(in)          :: d, m0,rm0
    ! Iter. for solver and plane search    
    integer,               intent(in)          :: niter, nAop, psiter  
    logical,               intent(in)          :: verb
    real,                  intent(in)          :: eps
    character(10)                              :: norm     ! Norm type to be used
    real,                  intent(in)          :: qnt      ! Quantile to set R_thr (R_bar)
    real, dimension(:),    intent(out)         :: m,err, resd,resm
    real, dimension(:,:),  intent(out)         ::        rmov,mmov
    real, dimension(size(d))                   :: rwd
    real, dimension(nAop)                      :: rwm
    real, dimension(size( m))                  :: g
    real, dimension(size( d) + nAop), target   :: rr, gg, tt
    real, dimension(:), pointer                :: rd, gd, td
    real, dimension(:), pointer                :: rm, gm, tm
    integer                                    :: iter, stat, psiter0
    logical                                    :: forget
    real                                       :: rt,rt1,rt2,quant  ! R_thr (R_bar), Quantile


    rd => rr(1:size(d))
    rm => rr(1+size(d):)
    gd => gg(1:size(d))
    gm => gg(1+size(d):)
    td => tt(1:size(d))
    tm => tt(1+size(d):)


    ! -------- begin initialization --------
    psiter0=1
    if (present(psiter)) psiter0=psiter

    if (present(Wop)) then
      stat=Wop(F,F,-d,rd) 
    else
      rd = -d                              !Rd  = -W  d
    end if 
    rm = 0.
    if (present(rm0)) then
      rm=rm0                               !Rm  =    Rm0
    end if
    if (present( m0)) then
      m=m0                                 !m   =     m0
      if (present(Wop)) then
        call chain0(Wop,Fop,F,T,m,rd,td)
      else
        stat=           Fop(F,T,m,rd   )   !Rd +=  WF m0
      end if 
      stat = Aop(F,T,eps*m0,rm)                           
      !Rm += e A m0
    else
      m=0
    end if 
    forget = T


    ! --------------- Set quantile and rt -------------
    if (.not. present(norm)) then
       norm="huber"
    end if

    quant=1.00
    if (present(qnt) .and. (qnt <= 1.0)) then
       quant = qnt;
    end if
    rt1=quantile( int(1.00 *size(rd)), abs(rd));
    rt2=quantile( int(quant*size(rd)), abs(rd));
    rt=rt1



    !---------------- begin iterations -----------
    do iter = 1,niter  

       !  --- Initialization ---
       if (iter == max(10,int(niter/10)) ) then 
          write (0,*) " ===> Rt1, Rt2 = ",rt1,rt2

          quant=0.50
          if (present(qnt) .and. (qnt <= 1.0)) then
             quant = qnt;
          end if
          rt1=quantile( int(quant*size(rwd)), abs(rwd));

          rt=rt2
          write (0,*) " ===> New Rt1, Rt (AFTER) = ",rt1,rt
       endif

       if (iter == 1)  then 
          call cgnorm_init(rt,norm,1)
       else
          call cgnorm_init(rt,norm,psiter0)
       endif
       !  --- -------------- ---


       ! Compute C'*Rd ==> g  = (WF)'*C'*Rd
       !         C'*Rm ==> g +=    A'*C'*Rm
       stat=g_norm(rwd,rd,norm,rt)

      if (present( Wop)) then
        call chain0(Wop,Fop,T,F,g,rwd,td)
      else
        stat =          Fop(T,F,g,rwd   )  ! g  = (WF)'Rwd
      end if 

      stat=g_norm(rwm,rm,norm,rt)
      stat =  Aop(T,T,g,eps*rwm)                           

      !g +=  e A'Rm
      if (present( Jop)) then
        tm=g
        stat=Jop(F,F,tm,g   )
      end if
      !g  =   J  g
      if (present( Wop)) then
        call chain0(Wop,Fop,F,F,g,gd,td)
      else
        stat =           Fop(F,F,g,gd   )
      end if 
      !Gd  = (WF) g
      stat =  Aop(F,F,eps*g,gm)                           
      !Gm  =  e A g



       ! Stepper run
       stat = stepper(forget, m,g, rr,gg)                  
       !m+=dm; R+=dR


      if (stat .eq.1) then
         ! got stuck descending
        write (0,*) "Stuck in descending ==> Exit !"
        exit
      end if


       ! ----------- report -----------
      if (present( mmov)) then
        mmov(:,iter) =  m(:size(mmov,1))
      end if
      if (present( rmov)) then
        rmov(:,iter) = rr(:size(rmov,1))
      end if
      if (present( err )) then
        err(  iter) = dot_product(rd,rd)
      end if
      if (present( verb)) then
        if (verb) then
          call solver_report(iter,m,g,rd,rm)
        end if
      end if
      forget=F
    end do 
    if (present( resd)) then
      resd = rd
    end if
    if (present( resm)) then
      resm = rm(:size(resm))
    end if
  end subroutine 



  !-----------------------------------------------
  !               Quantile function
  !-----------------------------------------------
  recursive function quantile(k,a) result( value)
    integer,             intent (in)  :: k          
    ! position in array
    real, dimension (:), intent (in)  :: a
    real                              :: value      
    ! output value of quantile
    integer                           :: j
    real                              :: ak
    ak = a( k)
    j = count( a < ak)  
    ! how many a(:) < ak
    if ( j >= k) then
       value = quantile( k, pack( a, a < ak))
    else
       j = count( a > ak) + k - size( a)
       if ( j > 0) then
          value = quantile( j, pack( a, a > ak))
       else
          value = ak
       end if
    end if
  end function

end module 
