module solver_smp_nrm_mod
! 0 = W (F J m - d)

  use sep
  use chain0_mod 
  use solver_report_mod
  use cgnorm_mod

  implicit none
  logical, parameter, private  :: T = .true., F = .false.

  contains
  subroutine solver_smp_nrm( m,d, Fop, stepper, niter , &
    &    norm,qnt,psiter,Wop,Jop,m0,err,resd,mmov,rmov,verb)
    optional :: norm,qnt,psiter,Wop,Jop,m0,err,resd,mmov,rmov,verb
    interface
!-------------------------- begin definitions -----------

      integer function Fop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function
      integer function Wop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function
      integer function Jop(adj,add,m,d)
        real::m(:),d(:)
        logical,intent(in)::adj,add
      end function
      integer function stepper(forget,m,g,rr,gg)
        real, dimension(:)  ::        m,g,rr,gg
        logical             :: forget
      end function
    end interface
    real, dimension(:),    intent(in)          :: d, m0
    integer,               intent(in)          :: niter, psiter ! Iter. for solver and plane search
    logical,               intent(in)          :: verb
    character(10)                              :: norm          ! Norm type to be used
    real,                  intent(in)          :: qnt           ! Quantile to set R_thr (R_bar)
    real, dimension(:),    intent(out)         :: m,err, resd
    real, dimension(:,:),  intent(out)         ::        rmov, mmov
    real, dimension(size(m))                   :: g
    real, dimension(size(d)),         target   :: rr, gg
    real, dimension(size(d)+size(m)), target   :: tt
    real, dimension(size(d))                   :: rwd
    real, dimension(:), pointer                :: rd, gd, td
    real, dimension(:), pointer                :: rm, gm, tm
    integer                                    :: iter, stat, psiter0
    logical                                    :: forget
    real                                       :: rt,rt1,rt2,quant    ! R_thr (R_bar), Quantile
   
    rd => rr(1:size(d))
    gd => gg(1:size(d))
    td => tt(1:size(d))
    tm => tt(1+size(d):)

    ! -------- begin initialization --------
    psiter0=1
    if (present(psiter)) psiter0=psiter

   if (present( Wop)) then
      stat=Wop(F,F,-d,rd)
   else
      rd = -d
      !Rd = -W  d
   end if
    
    if (present( m0)) then
      m=m0
      !m  =     m0

      if (present( Wop)) then
        call chain0(Wop,Fop,F,T,m,rd,td)
      else
        stat =           Fop(F,T,m,rd   )
      end if
      !Rd+=  WF m0

    else
      m=0
    end if
    forget = T

    ! ------------- Set Norm, Quantile and Rt -------------
    if (.not. present(norm)) then
       norm="huber"
    end if

    quant=1.00
    if (present(qnt) .and. (qnt <= 1.0)) then
       quant = qnt;
    end if
    rt1=quantile( int(1.00 *size(rd)), abs(rd));
    rt2=quantile( int(quant*size(rd)), abs(rd));
    rt=rt1

    !------------- begin iterations ------------
    do iter = 1,niter

       !  --- Initialization ---
       if (iter == max(10,int(niter/10)) ) then 
          write (0,*) " ===> Rt1, Rt2 = ",rt1,rt2

          quant=0.50
          if (present(qnt) .and. (qnt <= 1.0)) then
             quant = qnt;
          end if
          rt1=quantile( int(quant*size(rwd)), abs(rwd));

          rt=rt2
          write (0,*) " ===> New Rt1, Rt (AFTER) = ",rt1,rt
       endif

       if (iter == 1)  then 
          call cgnorm_init(rt,norm,1)
       else
          call cgnorm_init(rt,norm,psiter0)
       endif
       !  --- -------------- ---

       stat=g_norm(rwd,rd,norm,rt)
       ! Compute C'*Rd ==> g= (WF)'*C'*Rd


       if (present(Wop)) then
          call chain0(Wop,Fop,T,F,g,rwd,td)
       else
          stat =          Fop(T,F,g,rwd)
          !g  = (WF)'Rwd
       end if
       
       if (present(Jop)) then
          tm=g
          stat  = Jop(F,F,tm, g  )
       end if
       !g  =   J  g
       
       if (present(Wop)) then
          call chain0(Wop,Fop,F,F,g,gd,td)
       else
          stat =          Fop(F,F,g,gd   )
          !Gd = (WF) g 
       end if
       


       ! Stepper run
       stat = stepper(forget, m,g, rr,gg)
       !m+=dm; R+=dR

       
       if (stat .eq.1) then
          ! got stuck descending
          write (0,*) "Stuck in descending ==> Exit !"
          exit
       end if


       ! ----------- report -----------
       if (present( mmov)) then
          mmov(:,iter) =  m(:size(mmov,1))
       end if
       if (present( rmov)) then
          rmov(:,iter) = rd(:size(rmov,1))
       end if
       if (present( err )) then
          err(  iter) = dot_product(rd,rd)
       end if
       if (present( verb)) then
          if (verb) then
             call solver_report(iter,m,g,rd)
          end if
       end if
       forget=F
    end do
    if (present( resd)) then
       resd = rd
    end if
  end subroutine


  !-----------------------------------------------
  !               Quantile function
  !-----------------------------------------------
  recursive function quantile(k,a) result( value)
    integer,             intent (in)  :: k          
    ! position in array
    real, dimension (:), intent (in)  :: a
    real                              :: value      
    ! output value of quantile
    integer                           :: j
    real                              :: ak
    ak = a( k)
    j = count( a < ak)  
    ! how many a(:) < ak
    if ( j >= k) then
       value = quantile( k, pack( a, a < ak))
    else
       j = count( a > ak) + k - size( a)
       if ( j > 0) then
          value = quantile( j, pack( a, a > ak))
       else
          value = ak
       end if
    end if
  end function
  

end module
